#ifndef NACB_MAT4x4_H
#define NACB_MAT4x4_H

#include "vec4.h"
#include "vec3.h"
#include "mat3x3.h"
#include "nspace.h"

#ifdef __PY_PLUS_PLUS
#include <boost/python/tuple.hpp>
#include <boost/python/extract.hpp>
using boost::python::tuple;
using boost::python::extract;
#endif


#include <sstream>
NACB_BEGIN

class Quaternion;

class Mat4x4{
 public:
  double data[16];
  //
  Mat4x4(){}
  Mat4x4(const Vec3d & p,const Quaternion & q);

  void operator*=(const Mat4x4 & other){
    mult_right(other);
  }
  Mat4x4 operator*(const Mat4x4 & other) const{
    Mat4x4 res;
    for(int i=0;i<4;i++){
      for(int j=0;j<4;j++){
	double sum=0;
	for(int k=0;k<4;k++){
	  sum+=this->get(i,k)*other.get(k,j);
	}
	res(i,j)=sum;
      }
    }
    return res;
  }
  template <class T2>
    Vec3<T2> transformPoint(const Vec3<T2>  & pt) const{
    return Vec3<T2>(pt.x*data[0]+pt.y*data[1]+pt.z*data[2]+data[3],
		    pt.x*data[4]+pt.y*data[5]+pt.z*data[6]+data[7],
		    pt.x*data[8]+pt.y*data[9]+pt.z*data[10]+data[11]);
  }
  template <class T2>
  Vec3<T2> projTransformPoint(const Vec3<T2>  & pt) const{
    double x=pt.x*data[0]+pt.y*data[1]+pt.z*data[2]+data[3];
    double y=pt.x*data[4]+pt.y*data[5]+pt.z*data[6]+data[7];
    double z=pt.x*data[8]+pt.y*data[9]+pt.z*data[10]+data[11];
    double w=pt.x*data[12]+pt.y*data[13]+pt.z*data[14]+data[15];
    return Vec3<T2>(x/w,y/w,z/w);
  }
  double get(int i,int j) const{
    return data[4*i+j];
  }
  double & operator()(int i,int j){
    return data[4*i+j];
  }
  double & operator[](int ind){
    return data[ind];
  }
  Mat4x4 transpose() const{
    Mat4x4 mat;
    for(int i=0;i<4;i++){
      for(int j=0;j<4;j++){
	mat(i,j)=get(j,i);
      }
    }
    return mat;
  }
  Mat4x4 eucInv() const;
  Mat4x4 inverse() const;

  static Mat4x4 scale(double sx,double sy,double sz){
    Mat4x4 e=Mat4x4::eye();
    e[0]=sx;
    e[5]=sy;
    e[10]=sz;
    return e;
  }
  static Mat4x4 trans(const Vec3d & t){
    Mat4x4 e=Mat4x4::eye();
    e[3]=t.x;
    e[7]=t.y;
    e[11]=t.z;
    return e;
  }
  static Mat4x4 trans(double x,double y,double z){
    Mat4x4 e=Mat4x4::eye();
    e[3]=x;
    e[7]=y;
    e[11]=z;
    return e;
  }
  static Mat4x4 eye(){
    Mat4x4 e;
    e.zero();
    e.data[0]=e.data[5]=e.data[10]=e.data[15]=1;
    return e;
  }
  static Mat4x4 angleAxis(const Vec3d & axis,double t);
  void to_eye(){
    zero();
    data[0]=data[5]=data[10]=data[15]=1;
  }
  void zero(){
    memset(data,0,sizeof(double)*16);
  }
  static Mat4x4 rand(double n=0,double x=1){
    Mat4x4 r;
    for(int i=0;i<16;i++)
      r.data[i]=double(::rand())/double(RAND_MAX)*(x-n)+n;
    return r;
  }
  static Mat4x4 zeros(){
    Mat4x4 ret;
    ret.zero();
    return ret;
  }
  static Mat4x4 dx(double t){
    Mat4x4 m;
    m.zero();
    double ct=-sin(t);
    double st=cos(t);
    m(1,1)= ct;
    m(1,2)=-st;
    m(2,1)= st;
    m(2,2)= ct;
    return m;
  }
  static Mat4x4 dy(double t){
    Mat4x4 m;
    m.zero();
    double ct=-sin(t);
    double st=cos(t);
    m(0,0)= ct;
    m(0,2)= st;
    m(2,0)=-st;
    m(2,2)= ct;
    return m;
  }
  static Mat4x4 dz(double t){
    Mat4x4  m;
    m.zero();
    double ct=-sin(t);
    double st=cos(t);
    m(0,0)= ct;
    m(0,1)=-st;
    m(1,0)= st;
    m(1,1)= ct;
    return m;
  }

  static Mat4x4 rotx(double t){
    Mat4x4 m;
    m.to_eye();
    double ct=cos(t);
    double st=sin(t);
    m(1,1)= ct;
    m(1,2)=-st;
    m(2,1)= st;
    m(2,2)= ct;
    return m;
  }
  static Mat4x4 roty(double t){
    Mat4x4 m;
    m.to_eye();
    double ct=cos(t);
    double st=sin(t);
    m(0,0)= ct;
    m(0,2)= st;
    m(2,0)=-st;
    m(2,2)= ct;
    return m;
  }
  static Mat4x4 rotz(double t){
    Mat4x4  m;
    m.to_eye();
    double ct=cos(t);
    double st=sin(t);
    m(0,0)= ct;
    m(0,1)=-st;
    m(1,0)= st;
    m(1,1)= ct;
    return m;
  }

  //Just here for a little gain in speed over full mat mult
  void dx_right(double t){
    double ct=-sin(t),st=cos(t);
    for(int i=0;i<3;i++){
      double d1=get(i,1);
      double d2=get(i,2);
      
      operator()(i,1)= d1*ct+d2*st;
      operator()(i,2)=-d1*st+d2*ct;

      operator()(i,0)=0;
      operator()(i,3)=0;
    }
    operator()(3,0)=
      operator()(3,1)=
      operator()(3,2)=
      operator()(3,3)=0;
  }
  void dy_right(double t){
    double ct=-sin(t),st=cos(t);
    for(int i=0;i<3;i++){
      double d1=get(i,0);
      double d2=get(i,2);
      
      operator()(i,0)= d1*ct-d2*st;
      operator()(i,2)= d1*st+d2*ct;
      
      operator()(i,1)=0;
      operator()(i,3)=0;
    }
    operator()(3,0)=
      operator()(3,1)=
      operator()(3,2)=
      operator()(3,3)=0;
  }
  void dz_right(double t){
    double ct=-sin(t),st=cos(t);
    for(int i=0;i<3;i++){
      double d1=get(i,0);
      double d2=get(i,1);
      
      operator()(i,0)= d1*ct+d2*st;
      operator()(i,1)=-d1*st+d2*ct;

      operator()(i,2)=0;
      operator()(i,3)=0;
    }
    operator()(3,0)=
      operator()(3,1)=
      operator()(3,2)=
      operator()(3,3)=0;
  }

  //Just here for a little gain in speed over full mat mult
  void rotx_right(double t){
    double ct=cos(t),st=sin(t);
    for(int i=0;i<3;i++){
      double d1=get(i,1);
      double d2=get(i,2);
      
      operator()(i,1)= d1*ct+d2*st;
      operator()(i,2)=-d1*st+d2*ct;
    }
  }
  void roty_right(double t){
    double ct=cos(t),st=sin(t);
    for(int i=0;i<3;i++){
      double d1=get(i,0);
      double d2=get(i,2);
      
      operator()(i,0)= d1*ct-d2*st;
      operator()(i,2)= d1*st+d2*ct;
    }
  }
  void rotz_right(double t){
    double ct=cos(t),st=sin(t);
    for(int i=0;i<3;i++){
      double d1=get(i,0);
      double d2=get(i,1);
      
      operator()(i,0)= d1*ct+d2*st;
      operator()(i,1)=-d1*st+d2*ct;
    }
  }
  
  void trans_left(double x,double y,double z){
    operator()(0,3)+=x;
    operator()(1,3)+=y;
    operator()(2,3)+=z;
  }
  void trans_right(double x,double y,double z){
    double ox=get(0,0)*x+get(0,1)*y+get(0,2)*z;
    double oy=get(1,0)*x+get(1,1)*y+get(1,2)*z;
    double oz=get(2,0)*x+get(2,1)*y+get(2,2)*z;
    operator()(0,3)+=ox;
    operator()(1,3)+=oy;
    operator()(2,3)+=oz;
  }
  Vec4d operator*(const Vec4d & pt) const{
    return Vec4d(get(0,0)*pt.x+get(0,1)*pt.y+get(0,2)*pt.z+get(0,3)*pt.w,
		 get(1,0)*pt.x+get(1,1)*pt.y+get(1,2)*pt.z+get(1,3)*pt.w,
		 get(2,0)*pt.x+get(2,1)*pt.y+get(2,2)*pt.z+get(2,3)*pt.w,
		 get(3,0)*pt.x+get(3,1)*pt.y+get(3,2)*pt.z+get(3,3)*pt.w);
  }
  Mat4x4 operator+(const Mat4x4 & other) const{
    Mat4x4 res;
    for(int i=0;i<16;i++)
      res.data[i]=data[i]+other.data[i];
    return res;
  }
  Mat4x4 operator-(const Mat4x4 & other) const{
    Mat4x4 res;
    for(int i=0;i<16;i++)
      res.data[i]=data[i]-other.data[i];
    return res;
  }
  Mat4x4 operator*(const double d) const{
    Mat4x4 res;
    for(int i=0;i<16;i++)
      res.data[i]=data[i]*d;
    return res;
  }
  void operator=(const Mat4x4 & other){
    memcpy(data,other.data,sizeof(double)*16);
  }
  void mult_right(const Mat4x4 & other){
    Mat4x4 temp=operator*(other);
    operator=(temp);
  }
  double len() const{
    double sum=0;
    for(int i=0;i<16;i++)sum+=data[i]*data[i];
    return sqrt(sum);
  }
#ifdef USING_CAPSYS_MATRIX
  void fromMatrix(const Matrix & o){
    assert(o.m*o.n==16);
    memcpy(data,o.data,sizeof(double)*16);
  }
#endif

#ifdef __gl_h_
  void glMult() const{
    double gldata[16];
    
    for(int i=0;i<4;i++)
      for(int j=0;j<4;j++)
	gldata[4*j+i]=get(i,j);
    glMultMatrixd(gldata);
  }
  static Mat4x4 modelview(){
    Mat4x4 ret;
    glGetDoublev(GL_MODELVIEW_MATRIX,ret.data);
    return ret.transpose();
  }
  static Mat4x4 projection(){
    Mat4x4 ret;
    glGetDoublev(GL_PROJECTION_MATRIX,ret.data);
    return ret.transpose();
  }
#else
  void glMult() const{
    THROW_INCLUDE_GL;
  }
  static Mat4x4 projection(){
    THROW_INCLUDE_GL;
  }
  static Mat4x4 modelview(){
    THROW_INCLUDE_GL;
  }
#endif

  operator std::string () const{
    std::ostringstream os;
    print(os);
    return os.str();
  }
  void print(const char * name="") const;
  void print(std::ostream & os) const {
    for(int i=0;i<4;i++){
      for(int j=0;j<4;j++)
	os << get(i,j) << " ";
      os << "\n";
    }
  }

#ifdef __PY_PLUS_PLUS
  void python_set(const tuple & t,double d){
    operator()(extract<int>(t[0]),extract<int>(t[1]))=d;
  }
  double & operator[](const tuple & t){
    return operator()(extract<int>(t[0]),extract<int>(t[1]));
  }
#endif
};

std::ostream & operator<<(std::ostream & os,const Mat4x4 & mat);

NACB_END
#endif
