#ifndef NACB_QUATERNION_H
#define NACB_QUATERNION_H

#include <math.h>
#include "nspace.h"
#include "vec3.h"
#include "mat3x3.h"
#include "mat4x4.h"

#include <sstream>
#include <iostream>

NACB_BEGIN

class Quaternion{
public:
  double a;
  Vec3d  v;
  Quaternion(const Quaternion & q){
    operator=(q);
  }
  Quaternion(const Vec3d & v,double a){
    this->v=v;
    this->a=a;
  }
  Quaternion(double a=1,const Vec3d & v=Vec3d(0,0,0)){
    this->v=v;
    this->a=a;
  }
  static Quaternion rod_dx(const Vec3d & v);
  static Quaternion rod_dy(const Vec3d & v);
  static Quaternion rod_dz(const Vec3d & v);

  //assume rotation axis
  static Quaternion rod(const Vec3d & v=Vec3d(0,0,0)){
    double angle=v.len();
    if(angle<1e-12)return Quaternion(1,Vec3d(0,0,0));
    return Quaternion(cos(angle/2),v*(sin(angle/2)/angle));
  }
  operator std::string () const{
    std::ostringstream os;
    print(os);
    return os.str();
  }
  void print(std::ostream & o)const{
    printf("[%f %f %f %f]",a,v.x,v.y,v.z);
  }
  void print(char * name=0){
    printf("%s %f,%f,%f,%f (%f)\n",name,a,v.x,v.y,v.z,len());
  }
  double dot(const Quaternion & q)const{
    return a*q.a+v.dot(q.v);
  }
  double len(){
    return sqrt(a*a+v.dot(v));
  }
  Vec3d rotateNormal(const Vec3d & n){
    Quaternion q=times(Quaternion(0,n)).times(conj());
    return q.v;
  }
  Vec3d rotate(const Vec3d & pt) const{
    Quaternion q=times(Quaternion(1,pt)).times(conj());
    return q.v;
  }
  Quaternion conj() const{
    return Quaternion(v*-1.0,a);
  }
  void operator*=(double d){
    a*=d;
    v*=d;
  }
  void operator/=(double d){
    a/=d;
    v/=d;
  }
  double normalize(){
    double l=len();
    if(fabs(l-1)>1e-13)operator/=(l);
    return l;
  }
  void operator*=(const Quaternion & q){
    double ta=a*q.a-v.dot(q.v);

    v=q.v*a+v*q.a+(v^q.v);
    a=ta;
    normalize();
  }
  Quaternion operator+(const Quaternion & q) const{
    return Quaternion(a+q.a,v+q.v);
  }
  Quaternion times(const Quaternion & q) const{
    return Quaternion(a*q.a-v.dot(q.v),
		      q.v*a+v*q.a+(v^q.v));
  }
  Quaternion operator*(const Quaternion & q) const{
    Quaternion ret(a*q.a-v.dot(q.v),
		 q.v*a+v*q.a+(v^q.v));
    ret.normalize();
    return ret;
  }
  void operator=(const Quaternion & q){
    this->a=q.a;
    this->v=q.v;
  }
#ifdef __gl_h_
  void glRotate() const{
    double angle=acos(a)*2;
    if(fabs(angle)<1e-12)return;
    Vec3d axis=v;
    axis.normalize();
    glRotated(180.0*angle/M_PI,axis.x,axis.y,axis.z);
  }
#else
  void glRotate() const{
    throw "Exception: must include <GL/gl.h> before using GL functions.";
  }
#endif

  static Quaternion slerp(double t,Quaternion & q1,Quaternion & q2){
    if(t<0)t=0;
    if(t>1)t=1;
    
    Quaternion o=q2;
    double ct=q1.dot(q2);
    if(ct<0){
      ct=-ct;
      o*=-1;
    }
    double st=sqrt(1-ct*ct);
    double theta=atan2(st,ct);
    if(st<1e-10){
      if(t<0.5)return q1;
      else return q2;
    }
    return q1*(sin(theta*(1.0-t))/st)+o*(sin(theta*t)/st);
    
  }
  static Quaternion fromMatrixRows(const Vec3d & r1,
				   const Vec3d & r2,
				   const Vec3d & r3);
  static Quaternion fromMatrix(const Mat3x3 & mat){
    return Quaternion::fromMatrixRows(Vec3d(mat.get(0,0),
					    mat.get(0,1),
					    mat.get(0,2)),
				      Vec3d(mat.get(1,0),
					    mat.get(1,1),
					    mat.get(1,2)),
				      Vec3d(mat.get(2,0),
					    mat.get(2,1),
					    mat.get(2,2)));
  }
  static Quaternion fromMatrix(const Mat4x4 & mat){
    return Quaternion::fromMatrixRows(Vec3d(mat.get(0,0),
					    mat.get(0,1),
					    mat.get(0,2)),
				      Vec3d(mat.get(1,0),
					    mat.get(1,1),
					    mat.get(1,2)),
				      Vec3d(mat.get(2,0),
					    mat.get(2,1),
					    mat.get(2,2)));
  }
  //FIXME:
  void getMatrix(Mat3x3 & mat)const;

  //FIXME:
  void getMatrixRows(Vec3d & r0,Vec3d & r1,Vec3d & r2)const{
    r0=Vec3d(1.0-2.0*(v.y*v.y+v.z*v.z),
	     2.0*(v.x*v.y-a*v.z),
	     2.0*(v.x*v.z+a*v.y));
    r1=Vec3d(2.0*(v.x*v.y+a*v.z),
	     1.0-2.0*(v.x*v.x+v.z*v.z),
	     2.0*(v.y*v.z-a*v.x));
    r2=Vec3d(2.0*(v.x*v.z-a*v.y),
	     2.0*(v.y*v.z+a*v.x),
	     1.0-2.0*(v.x*v.x+v.y*v.y));
  }
  //FIXME:
  void getMatrixCols(Vec3d & c0,Vec3d & c1,Vec3d & c2)const{
    c0=Vec3d(1.0-2.0*(v.y*v.y+v.z*v.z),
	     2.0*(v.x*v.y+a*v.z),
	     2.0*(v.x*v.z-a*v.y));
    c1=Vec3d(2.0*(v.x*v.y-a*v.z),
	     1.0-2.0*(v.x*v.x+v.z*v.z),
	     2.0*(v.y*v.z+a*v.x));
    c2=Vec3d(2.0*(v.x*v.z+a*v.y),
	     2.0*(v.y*v.z-a*v.x),
	     1.0-2.0*(v.x*v.x+v.y*v.y));
    /*
    glMatrixMode(GL_MODELVIEW_MATRIX);
    glPushMatrix();
    glLoadIdentity();
    glRotate();
    double mat[4][4];
    glGetDoublev(GL_MODELVIEW_MATRIX,(double *)mat);
    c1=Vec3d(mat[0][0],mat[0][1],mat[0][2]);
    c2=Vec3d(mat[1][0],mat[1][1],mat[1][2]);
    c3=Vec3d(mat[2][0],mat[2][1],mat[2][2]);
    glPopMatrix();*/
  }
  //doesn't necessarily work for zxz
  //order={0,1,2} ==> rz(ang[2])*ry(ang[1])*rx(ang[0])
  //Algorithm from pdf linked at http://www.euclideanspace.com/maths/geometry/rotations/conversions/quaternionToEuler/index.htm by Amy de Buitlir
  void toEuler(const char * order,double angles[3]) const;
};

std::ostream & operator<<(std::ostream & os,const Quaternion & q);

NACB_END

#endif // 
